#include <vtkm/worklet/DispatcherMapField.h>
#include <vtkm/worklet/WorkletMapField.h>

#include <vtkm/cont/ArrayHandle.h>
#include <vtkm/cont/VariantArrayHandle.h>

#include <vtkm/VecTraits.h>
#include <vtkm/VectorAnalysis.h>

#include <vtkm/io/reader/VTKDataSetReader.h>
#include <vtkm/io/writer/VTKDataSetWriter.h>
#include <vtkm/filter/Gradient.h>

namespace vtkm
{
namespace worklet
{

struct Magnitude
{
  class ComputeMagnitude : public vtkm::worklet::WorkletMapField
  {
  public:
    using ControlSignature = void(FieldIn inputVectors, FieldOut outputMagnitudes);
    using ExecutionSignature = _2(_1);

    using InputDomain = _1;

    template<typename T, vtkm::IdComponent Size>
    VTKM_EXEC T operator()(const vtkm::Vec<T, Size>& inVector) const
    {
      return vtkm::Magnitude(inVector);
    }
  };

  template<typename ValueType, typename Storage>
  VTKM_CONT static vtkm::cont::ArrayHandle<
    typename vtkm::VecTraits<ValueType>::ComponentType>
  Run(const vtkm::cont::ArrayHandle<ValueType, Storage>& input)
  {
    using ComponentType = typename vtkm::VecTraits<ValueType>::ComponentType;
    vtkm::cont::ArrayHandle<ComponentType> output;

    vtkm::worklet::DispatcherMapField<ComputeMagnitude> dispatcher;
    dispatcher.Invoke(input, output);

    return output;
  }
};

} // namespace worklet
} // namespace vtkm

#include <vtkm/filter/FilterField.h>

#include <vtkm/filter/internal/CreateResult.h>

namespace vtkm
{
namespace filter
{

class FieldMagnitude : public vtkm::filter::FilterField<FieldMagnitude>
{
public:
  VTKM_CONT
  FieldMagnitude();

  template<typename ArrayHandleType, typename Policy>
  VTKM_CONT vtkm::cont::DataSet DoExecute(
    const vtkm::cont::DataSet& inDataSet,
    const ArrayHandleType& inField,
    const vtkm::filter::FieldMetadata& fieldMetadata,
    vtkm::filter::PolicyBase<Policy>);
};

template<>
class FilterTraits<vtkm::filter::FieldMagnitude>
{
public:
  struct InputFieldTypeList : vtkm::ListTagBase<vtkm::Vec<vtkm::Float32, 2>,
                                                vtkm::Vec<vtkm::Float64, 2>,
                                                vtkm::Vec<vtkm::Float32, 3>,
                                                vtkm::Vec<vtkm::Float64, 3>,
                                                vtkm::Vec<vtkm::Float32, 4>,
                                                vtkm::Vec<vtkm::Float64, 4>>
  {
  };
};

VTKM_CONT
FieldMagnitude::FieldMagnitude()
{
  this->SetOutputFieldName("");
}

template<typename ArrayHandleType, typename Policy>
VTKM_CONT cont::DataSet FieldMagnitude::DoExecute(
  const vtkm::cont::DataSet& inDataSet,
  const ArrayHandleType& inField,
  const vtkm::filter::FieldMetadata& fieldMetadata,
  vtkm::filter::PolicyBase<Policy>)
{
  VTKM_IS_ARRAY_HANDLE(ArrayHandleType);

  using ComponentType =
    typename vtkm::VecTraits<typename ArrayHandleType::ValueType>::ComponentType;

  vtkm::cont::ArrayHandle<ComponentType> outField =
    vtkm::worklet::Magnitude::Run(inField);

  std::string outFieldName = this->GetOutputFieldName();
  if (outFieldName == "")
  {
    outFieldName = fieldMetadata.GetName() + "_magnitude";
  }

  return vtkm::filter::internal::CreateResult(inDataSet,
                                              outField,
                                              outFieldName,
                                              fieldMetadata.GetAssociation(),
                                              fieldMetadata.GetCellSetName());
}

} // namespace filter
} // namespace vtkm


int main()
{
  const char *input = "kitchen.vtk";
  vtkm::io::reader::VTKDataSetReader reader(input);
  vtkm::cont::DataSet ds_from_file = reader.ReadDataSet();

  vtkm::filter::Gradient grad;
  grad.SetActiveField("c1");
  vtkm::cont::DataSet ds_from_grad = grad.Execute(ds_from_file);

  vtkm::filter::FieldMagnitude mag;
  mag.SetActiveField("Gradients");
  vtkm::cont::DataSet mag_grad = grad.Execute(ds_from_grad);
  
  vtkm::io::writer::VTKDataSetWriter writer("out_mag_grad.vtk");
  writer.WriteDataSet(mag_grad);

  return 0;
}
